@@ -19,9 +19,12 @@ it under the terms of the one of three licenses as you choose:
 
  */
 
+#include <math.h>
 #include <errno.h>
 #include <float.h>
-#include <math.h>
+#include <new>
+#include <sys/types.h>
+#include <sys/stat.h>
 #ifndef WIN32
 #include <netinet/in.h>
 #else
@@ -29,6 +32,7 @@ it under the terms of the one of three licenses as you choose:
 #endif
 #define LIBRAW_LIBRARY_BUILD
 #include "libraw/libraw.h"
+//#include "internal/defines.h"
 
 #ifdef __cplusplus
 extern "C" 
@@ -75,6 +79,8 @@ extern "C"
                 return "Input/output error";
             case LIBRAW_CANCELLED_BY_CALLBACK:
                 return "Cancelled by user callback";
+            case LIBRAW_BAD_CROP:
+                return "Bad crop box";
             default:
                 return "Unknown error code";
         }
@@ -104,7 +110,7 @@ const float LibRaw_constants::d65_white[3] =  { 0.950456, 1, 1.088754 };
 #define ID libraw_internal_data.internal_data
 
 #define EXCEPTION_HANDLER(e) do{                        \
-        fprintf(stderr,"Exception %d caught\n",e);      \
+        /* fprintf(stderr,"Exception %d caught\n",e);*/ \
         switch(e)                                       \
             {                                           \
             case LIBRAW_EXCEPTION_ALLOC:                \
@@ -121,6 +127,9 @@ const float LibRaw_constants::d65_white[3] =  { 0.950456, 1, 1.088754 };
             case LIBRAW_EXCEPTION_CANCELLED_BY_CALLBACK:\
                 recycle();                              \
                 return LIBRAW_CANCELLED_BY_CALLBACK;    \
+            case LIBRAW_EXCEPTION_BAD_CROP:             \
+                recycle();                              \
+                return LIBRAW_BAD_CROP;                 \
             default:                                    \
                 return LIBRAW_UNSPECIFIED_ERROR;        \
             } \
@@ -152,6 +161,11 @@ void LibRaw::derror()
     libraw_internal_data.unpacker_data.data_error++;
 }
 
+void LibRaw::dcraw_clear_mem(libraw_processed_image_t* p)
+{
+    if(p) ::free(p);
+}
+
 #define ZERO(a) memset(&a,0,sizeof(a))
 
 
@@ -160,6 +174,7 @@ LibRaw:: LibRaw(unsigned int flags)
     double aber[4] = {1,1,1,1};
     double gamm[6] = { 0.45,4.5,0,0,0,0 };
     unsigned greybox[4] =  { 0, 0, UINT_MAX, UINT_MAX };
+    unsigned cropbox[4] =  { 0, 0, UINT_MAX, UINT_MAX };
 #ifdef DCRAW_VERBOSE
     verbose = 1;
 #else
@@ -173,6 +188,7 @@ LibRaw:: LibRaw(unsigned int flags)
     memmove(&imgdata.params.aber,&aber,sizeof(aber));
     memmove(&imgdata.params.gamm,&gamm,sizeof(gamm));
     memmove(&imgdata.params.greybox,&greybox,sizeof(greybox));
+    memmove(&imgdata.params.cropbox,&cropbox,sizeof(cropbox));
     
     imgdata.params.bright=1;
     imgdata.params.use_camera_matrix=-1;
@@ -183,8 +199,10 @@ LibRaw:: LibRaw(unsigned int flags)
     imgdata.params.output_color=1;
     imgdata.params.output_bps=8;
     imgdata.params.use_fuji_rotate=1;
+    imgdata.params.exp_shift = 1.0;
     imgdata.params.auto_bright_thr = LIBRAW_DEFAULT_AUTO_BRIGHTNESS_THRESHOLD;
     imgdata.params.adjust_maximum_thr= LIBRAW_DEFAULT_ADJUST_MAXIMUM_THRESHOLD;
+    imgdata.params.green_matching = 0;
     imgdata.parent_class = this;
     imgdata.progress_flags = 0;
     tls = new LibRaw_TLS;
@@ -197,6 +215,13 @@ void* LibRaw:: malloc(size_t t)
     void *p = memmgr.malloc(t);
     return p;
 }
+void* LibRaw:: realloc(void *q,size_t t)
+{
+    void *p = memmgr.realloc(q,t);
+    return p;
+}
+
+
 void* LibRaw::       calloc(size_t n,size_t t)
 {
     void *p = memmgr.calloc(n,t);
@@ -254,7 +279,7 @@ void LibRaw:: recycle()
     ZERO(imgdata.masked_pixels);
     ZERO(imgdata.sizes);
     ZERO(imgdata.color);
-    ZERO(libraw_internal_data.internal_output_params);
+    ZERO(libraw_internal_data);
     memmgr.cleanup();
     imgdata.thumbnail.tformat = LIBRAW_THUMBNAIL_UNKNOWN;
     imgdata.progress_flags = 0;
@@ -275,6 +300,7 @@ const char * LibRaw::unpack_function_name()
     if (load_raw == &LibRaw::canon_sraw_load_raw)       return "canon_sraw_load_raw()"; //+
 
     if (load_raw == &LibRaw::eight_bit_load_raw )       return "eight_bit_load_raw()"; //+
+    if (load_raw == &LibRaw::foveon_load_raw )          return "foveon_load_raw()";
     if (load_raw == &LibRaw::fuji_load_raw )            return "fuji_load_raw()"; //+
     // 10
     if (load_raw == &LibRaw::hasselblad_load_raw )      return "hasselblad_load_raw()"; //+
@@ -461,7 +487,7 @@ int LibRaw::add_masked_borders_to_bitmap()
     if(S.width != S.iwidth || S.height!=S.iheight)
         return LIBRAW_CANNOT_ADDMASK;
 
-    if(!P1.filters)
+    if(P1.is_foveon || !P1.filters)
         return LIBRAW_CANNOT_ADDMASK;
         
     if(!imgdata.image)
@@ -500,7 +526,7 @@ int LibRaw::add_masked_borders_to_bitmap()
             for(c=S.left_margin; c<S.left_margin+S.iwidth;c++)
                 {
                     int col = c - S.left_margin;
-                    newimage[r*S.raw_width+c][COLOR(r,c)] = imgdata.image[row*S.iwidth+col][COLOR(r,c)];
+                    newimage[r*S.raw_width+c][COLOR(r,c)] = imgdata.image[row*S.iwidth+col][COLOR(row,col)];
 //                    for(int cc=0;cc<4;cc++)
 //                        newimage[r*S.raw_width+c][cc] = imgdata.image[row*S.iwidth+col][cc];
                 }
@@ -526,10 +552,33 @@ int LibRaw::add_masked_borders_to_bitmap()
     return LIBRAW_SUCCESS;
 }
 
-int LibRaw::open_file(const char *fname)
+int LibRaw::open_file(const char *fname, INT64 max_buf_size)
 {
-    // this stream will close on recycle()
-    LibRaw_file_datastream *stream = new LibRaw_file_datastream(fname);
+#ifndef WIN32
+    struct stat st;
+    if(stat(fname,&st))
+        return LIBRAW_IO_ERROR;
+    int big = (st.st_size > max_buf_size)?1:0;
+#else
+	struct _stati64 st;
+    if(_stati64(fname,&st))	
+        return LIBRAW_IO_ERROR;
+    int big = (st.st_size > max_buf_size)?1:0;
+#endif
+
+    LibRaw_abstract_datastream *stream;
+    try {
+        if(big)
+         stream = new LibRaw_bigfile_datastream(fname);
+        else
+         stream = new LibRaw_file_datastream(fname);
+    }
+
+    catch (std::bad_alloc)
+        {
+            recycle();
+            return LIBRAW_UNSUFFICIENT_MEMORY;
+        }
     if(!stream->valid())
         {
             delete stream;
@@ -555,7 +604,15 @@ int LibRaw::open_buffer(void *buffer, size_t size)
     if(!buffer  || buffer==(void*)-1)
         return LIBRAW_IO_ERROR;
 
-    LibRaw_buffer_datastream *stream = new LibRaw_buffer_datastream(buffer,size);
+    LibRaw_buffer_datastream *stream;
+    try {
+        stream = new LibRaw_buffer_datastream(buffer,size);
+    }
+    catch (std::bad_alloc)
+        {
+            recycle();
+            return LIBRAW_UNSUFFICIENT_MEMORY;
+        }
     if(!stream->valid())
         {
             delete stream;
@@ -598,7 +655,7 @@ int LibRaw::open_datastream(LibRaw_abstract_datastream *stream)
             {
                 IO.fwidth = S.width;
                 IO.fheight = S.height;
-                S.iwidth = S.width = IO.fuji_width << !libraw_internal_data.unpacker_data.fuji_layout;
+                S.iwidth = S.width = IO.fuji_width << (int)(!libraw_internal_data.unpacker_data.fuji_layout);
                 S.iheight = S.height = S.raw_height;
                 S.raw_height += 2*S.top_margin;
             }
@@ -661,7 +718,9 @@ int LibRaw::open_datastream(LibRaw_abstract_datastream *stream)
             S.width  += S.width  & 1;
         }
 
-    IO.shrink = P1.filters && (O.half_size || O.threshold || O.aber[0] != 1 || O.aber[2] != 1);
+    IO.shrink = P1.filters && (O.half_size ||
+	((O.threshold || O.aber[0] != 1 || O.aber[2] != 1) ));
+
     S.iheight = (S.height + IO.shrink) >> IO.shrink;
     S.iwidth  = (S.width  + IO.shrink) >> IO.shrink;
     
@@ -723,7 +782,16 @@ int LibRaw::unpack(void)
 
         if (O.filtering_mode & LIBRAW_FILTERING_AUTOMATIC_BIT)
             O.filtering_mode = LIBRAW_FILTERING_AUTOMATIC; // restore automated mode
-        
+
+        // adjust black to possible maximum
+        unsigned int i = C.cblack[3];
+        unsigned int c;
+        for(c=0;c<3;c++)
+            if (i > C.cblack[c]) i = C.cblack[c];
+        for (c=0;c<4;c++)
+            C.cblack[c] -= i;
+        C.black += i;
+
         SET_PROC_FLAG(LIBRAW_PROGRESS_LOAD_RAW);
         RUN_CALLBACK(LIBRAW_PROGRESS_LOAD_RAW,1,2);
         
@@ -741,20 +809,38 @@ int LibRaw::dcraw_document_mode_processing(void)
 
     try {
 
-        if(IO.fwidth) 
-            rotate_fuji_raw();
-
         if(!own_filtering_supported() && (O.filtering_mode & LIBRAW_FILTERING_AUTOMATIC_BIT))
             O.filtering_mode = LIBRAW_FILTERING_AUTOMATIC_BIT; // turn on black and zeroes filtering
 
-        O.document_mode = 2;
-
-        O.use_fuji_rotate = 0;
         if (!(O.filtering_mode & LIBRAW_FILTERING_NOZEROES) && IO.zero_is_bad)
             {
                 remove_zeroes();
                 SET_PROC_FLAG(LIBRAW_PROGRESS_REMOVE_ZEROES);
             }
+
+        if (O.user_black >= 0) 
+            C.black = O.user_black;
+        subtract_black();
+
+        if(IO.fwidth) 
+            rotate_fuji_raw();
+
+        O.document_mode = 2;
+        
+        if(P1.is_foveon)
+            {
+                // filter image data for foveon document mode
+                short *iptr = (short *)imgdata.image;
+                for (int i=0; i < S.height*S.width*4; i++)
+                    {
+                        if ((short) iptr[i] < 0) 
+                            iptr[i] = 0;
+                    }
+                SET_PROC_FLAG(LIBRAW_PROGRESS_FOVEON_INTERPOLATE);
+            }
+
+        O.use_fuji_rotate = 0;
+
         if(O.bad_pixels) 
             {
                 bad_pixels(O.bad_pixels);
@@ -765,11 +851,10 @@ int LibRaw::dcraw_document_mode_processing(void)
                 subtract (O.dark_frame);
                 SET_PROC_FLAG(LIBRAW_PROGRESS_DARK_FRAME);
             }
-        if(O.filtering_mode & LIBRAW_FILTERING_NOBLACKS)
-            C.black=0;
+        if (~O.cropbox[2] && ~O.cropbox[3]) crop_pixels();
 
-        if (O.user_black >= 0) 
-            C.black = O.user_black;
+
+        adjust_maximum();
 
         if (O.user_sat > 0) 
             C.maximum = O.user_sat;
@@ -785,14 +870,14 @@ int LibRaw::dcraw_document_mode_processing(void)
             }
         SET_PROC_FLAG(LIBRAW_PROGRESS_MIX_GREEN);
 
-        if ( P1.colors == 3) 
+        if (!P1.is_foveon && P1.colors == 3) 
             median_filter();
         SET_PROC_FLAG(LIBRAW_PROGRESS_MEDIAN_FILTER);
 
-        if ( O.highlight == 2) 
+        if (!P1.is_foveon && O.highlight == 2) 
             blend_highlights();
 
-        if ( O.highlight > 2) 
+        if (!P1.is_foveon && O.highlight > 2) 
             recover_highlights();
         SET_PROC_FLAG(LIBRAW_PROGRESS_HIGHLIGHTS);
 
@@ -1266,6 +1351,14 @@ int LibRaw::unpack_thumb(void)
                         return 0;
 
                     }
+                else if (write_thumb == &LibRaw::foveon_thumb)
+                    {
+                        foveon_thumb_loader();
+                        // may return with error, so format is set in
+                        // foveon thumb loader itself
+                        SET_PROC_FLAG(LIBRAW_PROGRESS_THUMB_LOAD);
+                        return 0;
+                    }
                 // else if -- all other write_thumb cases!
                 else
                     {
@@ -1407,33 +1500,171 @@ int LibRaw::rotate_fuji_raw(void)
     
 }
 
+void LibRaw::subtract_black()
+{
 
+#define BAYERC(row,col,c) imgdata.image[((row) >> IO.shrink)*S.iwidth + ((col) >> IO.shrink)][c] 
+
+    if(imgdata.masked_pixels.ph1_black)
+        {
+            // Phase One compressed format
+            int row,col,val,cc;
+            for(row=0;row<S.height;row++)
+                for(col=0;col<S.width;col++)
+                    {
+                        cc=FC(row,col);
+                        val = BAYERC(row,col,cc) 
+                            - imgdata.color.phase_one_data.t_black 
+                            + imgdata.masked_pixels.ph1_black[row+S.top_margin][(col + S.left_margin) 
+                                                                                >=C.phase_one_data.split_col];
+                        if(val<0) val = 0;
+                        BAYERC(row,col,cc) = val;
+                    }
+            C.maximum -= C.black;
+            if(!( O.filtering_mode & LIBRAW_FILTERING_NORAWCURVE) )
+                {
+                    phase_one_correct();
+                }
+            // recalculate channel maximum
+            ZERO(C.channel_maximum);
+            for(row=0;row<S.height;row++)
+                for(col=0;col<S.width;col++)
+                    {
+                        cc=FC(row,col);
+                        val = BAYERC(row,col,cc);
+                        if(C.channel_maximum[cc] > val) C.channel_maximum[cc] = val;
+                    }
+            // clear P1 black level data
+            imgdata.color.phase_one_data.t_black = 0;
+            if(imgdata.masked_pixels.ph1_black)
+                {
+                    free(imgdata.masked_pixels.ph1_black);
+                    imgdata.masked_pixels.ph1_black = 0;
+                }
+            ZERO(C.cblack);
+            C.black = 0;
+        }
+    else if((C.black || C.cblack[0] || C.cblack[1] || C.cblack[2] || C.cblack[3]))
+        {
+            int cblk[4],i,row,col,val,cc;
+            for(i=0;i<4;i++)
+                cblk[i] = C.cblack[i]+C.black;
+            ZERO(C.channel_maximum);
+
+            for(row=0;row<S.height;row++)
+                for(col=0;col<S.width;col++)
+                    {
+                        cc=COLOR(row,col);
+                        val = BAYERC(row,col,cc);
+                        if(val > cblk[cc])
+                            val -= cblk[cc];
+                        else
+                            val = 0;
+                        if(C.channel_maximum[cc] < val) C.channel_maximum[cc] = val;
+                        BAYERC(row,col,cc) = val;
+                    }
+            C.maximum -= C.black;
+            ZERO(C.cblack);
+            C.black = 0;
+        }
+}
+
+#define TBLN 65535
+
+void LibRaw::exp_bef(float shift, float smooth)
+{
+    // params limits
+    if(shift>8) shift = 8;
+    if(shift<0.25) shift = 0.25;
+    if(smooth < 0.0) smooth = 0.0;
+    if(smooth > 1.0) smooth = 1.0;
+    
+    unsigned short *lut = (ushort*)malloc((TBLN+1)*sizeof(unsigned short));
+
+    if(shift <=1.0)
+        {
+            for(int i=0;i<=TBLN;i++)
+                lut[i] = ushort((float)i*shift);
+        }
+    else
+        {
+            float x1,x2,y1,y2;
+
+            float cstops = log(shift)/log(2.0f);
+            float room = cstops*2;
+            float roomlin = powf(2.0f,room);
+            x2 = (float)TBLN;
+            x1 = (x2+1)/roomlin-1;
+            y1 = x1*shift;
+            y2 = x2*(1+(1-smooth)*(shift-1));
+            float sq3x=powf(x1*x1*x2,1.0f/3.0f);
+            float B = (y2-y1+shift*(3*x1-3.0f*sq3x)) / (x2+2.0f*x1-3.0f*sq3x);
+            float A = (shift - B)*3.0f*powf(x1*x1,1.0f/3.0f);
+            float CC = y2 - A*powf(x2,1.0f/3.0f)-B*x2;
+            for(int i=0;i<=TBLN;i++)
+                {
+                    float X = (float)i;
+                    float Y = A*powf(X,1.0f/3.0f)+B*X+CC;
+                    if(i<x1)
+                        lut[i] = ushort((float)i*shift);
+                    else
+                        lut[i] = Y<0?0:(Y>TBLN?TBLN:ushort(Y));
+                }
+        }
+    for(int i=0; i< S.height*S.width; i++)
+        {
+            imgdata.image[i][0] = lut[imgdata.image[i][0]];
+            imgdata.image[i][1] = lut[imgdata.image[i][1]];
+            imgdata.image[i][2] = lut[imgdata.image[i][2]];
+            imgdata.image[i][3] = lut[imgdata.image[i][3]];
+        }
+    for(int i=0;i<4;i++)
+        C.channel_maximum[i] = lut[C.channel_maximum[i]];
+    C.maximum = lut[C.maximum];
+    // no need to adjust the minumum, black is already subtracted
+    free(lut);
+}
 int LibRaw::dcraw_process(void)
 {
     int quality,i;
 
+    int iterations=-1, dcb_enhance=1, noiserd=0;
+    int eeci_refine_fl=0, es_med_passes_fl=0;
+    float cared=0,cablue=0;
+    float linenoise=0; 
+    float lclean=0,cclean=0;
+    float thresh=0;
+    float preser=0;
+    float expos=1.0;
+
 
     CHECK_ORDER_LOW(LIBRAW_PROGRESS_LOAD_RAW);
     CHECK_ORDER_HIGH(LIBRAW_PROGRESS_PRE_INTERPOLATE);
 
     try {
-
-        adjust_maximum();
-        if(IO.fwidth) 
-            rotate_fuji_raw();
-
+        
+        int save_4color = O.four_color_rgb;
 
         if(!own_filtering_supported() && (O.filtering_mode & LIBRAW_FILTERING_AUTOMATIC_BIT))
             O.filtering_mode = LIBRAW_FILTERING_AUTOMATIC_BIT; // turn on black and zeroes filtering
 
-        if(O.half_size) 
-            O.four_color_rgb = 1;
-
         if (!(O.filtering_mode & LIBRAW_FILTERING_NOZEROES) && IO.zero_is_bad) 
             {
                 remove_zeroes();
                 SET_PROC_FLAG(LIBRAW_PROGRESS_REMOVE_ZEROES);
             }
+
+        // black already set to possible maximum
+        if (O.user_black >= 0) C.black = O.user_black;
+        subtract_black();
+
+        if(IO.fwidth) 
+            rotate_fuji_raw();
+
+
+        if(O.half_size) 
+            O.four_color_rgb = 1;
+
         if(O.bad_pixels) 
             {
                 bad_pixels(O.bad_pixels);
@@ -1445,35 +1676,87 @@ int LibRaw::dcraw_process(void)
                 SET_PROC_FLAG(LIBRAW_PROGRESS_DARK_FRAME);
             }
 
+        if (~O.cropbox[2] && ~O.cropbox[3]) crop_pixels();
+
         quality = 2 + !IO.fuji_width;
 
-        if(O.filtering_mode & LIBRAW_FILTERING_NOBLACKS)
-            C.black=0;
-
         if (O.user_qual >= 0) quality = O.user_qual;
-        if (O.user_black >= 0) C.black = O.user_black;
+
+        adjust_maximum();
+
         if (O.user_sat > 0) C.maximum = O.user_sat;
 
+        if (P1.is_foveon && !O.document_mode) 
+            {
+                foveon_interpolate();
+                SET_PROC_FLAG(LIBRAW_PROGRESS_FOVEON_INTERPOLATE);
+            }
 
-        if ( O.document_mode < 2)
+        if (O.green_matching && !O.half_size)
+            {
+                green_matching();
+            }
+
+        if (!P1.is_foveon &&  O.document_mode < 2)
             {
                 scale_colors();
                 SET_PROC_FLAG(LIBRAW_PROGRESS_SCALE_COLORS);
             }
 
         pre_interpolate();
+
         SET_PROC_FLAG(LIBRAW_PROGRESS_PRE_INTERPOLATE);
 
+        if (O.dcb_iterations >= 0) iterations = O.dcb_iterations;
+        if (O.dcb_enhance_fl >=0 ) dcb_enhance = O.dcb_enhance_fl;
+        if (O.fbdd_noiserd >=0 ) noiserd = O.fbdd_noiserd;
+        if (O.eeci_refine >=0 ) eeci_refine_fl = O.eeci_refine;
+        if (O.es_med_passes >0 ) es_med_passes_fl = O.es_med_passes;
+
+// LIBRAW_DEMOSAIC_PACK_GPL3
+
+        if (!O.half_size && O.cfa_green >0) {thresh=O.green_thresh ;green_equilibrate(thresh);} 
+        if (O.exp_correc >0) {expos=O.exp_shift ; preser=O.exp_preser; exp_bef(expos,preser);} 
+        if (O.ca_correc >0 ) {cablue=O.cablue; cared=O.cared; CA_correct_RT(cablue, cared);}
+        if (O.cfaline >0 ) {linenoise=O.linenoise; cfa_linedn(linenoise);}
+        if (O.cfa_clean >0 ) {lclean=O.lclean; cclean=O.cclean; cfa_impulse_gauss(lclean,cclean);}
+
         if (P1.filters && !O.document_mode) 
             {
+                if (noiserd>0 && P1.colors==3 && P1.filters) fbdd(noiserd);
+
                 if (quality == 0)
                     lin_interpolate();
                 else if (quality == 1 || P1.colors > 3)
                     vng_interpolate();
                 else if (quality == 2)
                     ppg_interpolate();
-                else 
+
+                else if (quality == 3) 
+                    ahd_interpolate(); // really don't need it here due to fallback op
+
+                else if (quality == 4)
+                    dcb(iterations, dcb_enhance);
+
+//  LIBRAW_DEMOSAIC_PACK_GPL2                
+                else if (quality == 5)
+                    ahd_interpolate_mod();
+                else if (quality == 6)
+                    afd_interpolate_pl(2,1);
+                else if (quality == 7)
+                    vcd_interpolate(0);
+                else if (quality == 8)
+                    vcd_interpolate(12);
+                else if (quality == 9)
+                    lmmse_interpolate(1);
+
+// LIBRAW_DEMOSAIC_PACK_GPL3
+                else if (quality == 10)
+                    amaze_demosaic_RT();
+ // fallback to AHD
+                else
                     ahd_interpolate();
+                
                 SET_PROC_FLAG(LIBRAW_PROGRESS_INTERPOLATE);
             }
         if (IO.mix_green)
@@ -1483,10 +1766,22 @@ int LibRaw::dcraw_process(void)
                 SET_PROC_FLAG(LIBRAW_PROGRESS_MIX_GREEN);
             }
 
-        if (P1.colors == 3) 
+        if(!P1.is_foveon)
             {
-                median_filter();
-                SET_PROC_FLAG(LIBRAW_PROGRESS_MEDIAN_FILTER);
+                if (P1.colors == 3) 
+                    {
+                        
+                        if (quality == 8) 
+                            {
+                                if (eeci_refine_fl == 1) refinement();
+                                if (O.med_passes > 0)    median_filter_new();
+                                if (es_med_passes_fl > 0) es_median_filter();
+                            } 
+                        else {
+                            median_filter();
+                        }
+                        SET_PROC_FLAG(LIBRAW_PROGRESS_MEDIAN_FILTER);
+                    }
             }
         
         if (O.highlight == 2) 
@@ -1530,6 +1825,9 @@ int LibRaw::dcraw_process(void)
             }
         if (O.filtering_mode & LIBRAW_FILTERING_AUTOMATIC_BIT)
             O.filtering_mode = LIBRAW_FILTERING_AUTOMATIC; // restore automated mode
+
+        O.four_color_rgb = save_4color; // also, restore
+
         return 0;
     }
     catch ( LibRaw_exceptions err) {
@@ -1578,6 +1876,7 @@ static const char  *static_camera_list[] =
 "Canon PowerShot G9",
 "Canon PowerShot G10",
 "Canon PowerShot G11",
+"Canon PowerShot G12",
 "Canon PowerShot S2 IS (CHDK hack)",
 "Canon PowerShot S3 IS (CHDK hack)",
 "Canon PowerShot S5 IS (CHDK hack)",
@@ -1589,8 +1888,11 @@ static const char  *static_camera_list[] =
 "Canon PowerShot S60",
 "Canon PowerShot S70",
 "Canon PowerShot S90",
+"Canon PowerShot S95",
 "Canon PowerShot SX1 IS",
 "Canon PowerShot SX110 IS (CHDK hack)",
+"Canon PowerShot SX120 IS (CHDK hack)",
+"Canon PowerShot SX20 IS (CHDK hack)",
 "Canon EOS D30",
 "Canon EOS D60",
 "Canon EOS 5D",
@@ -1601,11 +1903,13 @@ static const char  *static_camera_list[] =
 "Canon EOS 30D",
 "Canon EOS 40D",
 "Canon EOS 50D",
+"Canon EOS 60D",
 "Canon EOS 300D / Digital Rebel / Kiss Digital",
 "Canon EOS 350D / Digital Rebel XT / Kiss Digital N",
 "Canon EOS 400D / Digital Rebel XTi / Kiss Digital X",
 "Canon EOS 450D / Digital Rebel XSi / Kiss Digital X2",
 "Canon EOS 500D / Digital Rebel T1i / Kiss Digital X3",
+"Canon EOS 550D / Digital Rebel T2i / Kiss Digital X4",
 "Canon EOS 1000D / Digital Rebel XS / Kiss Digital F",
 "Canon EOS D2000C",
 "Canon EOS-1D",
@@ -1633,6 +1937,8 @@ static const char  *static_camera_list[] =
 "Casio EX-Z75",
 "Casio EX-Z750",
 "Casio EX-Z850",
+"Casio EX-Z1050",
+"Casio EX-Z1080",
 "Casio Exlim Pro 505",
 "Casio Exlim Pro 600",
 "Casio Exlim Pro 700",
@@ -1659,16 +1965,18 @@ static const char  *static_camera_list[] =
 "Fuji FinePix S9000/S9500",
 "Fuji FinePix S9100/S9600",
 "Fuji FinePix S200EXR",
+"Fuji FinePix HS10/HS11",
 "Fuji IS-1",
 "Hasselblad CFV",
 "Hasselblad H3D",
+"Hasselblad H4D",
 "Hasselblad V96C",
 "Imacon Ixpress 16-megapixel",
 "Imacon Ixpress 22-megapixel",
 "Imacon Ixpress 39-megapixel",
 "ISG 2020x1520",
-"Kodak DC20 (see Oliver Hartman's page)",
-"Kodak DC25 (see Jun-ichiro Itoh's page)",
+"Kodak DC20",
+"Kodak DC25",
 "Kodak DC40",
 "Kodak DC50",
 "Kodak DC120 (also try kdc2tiff)",
@@ -1703,6 +2011,7 @@ static const char  *static_camera_list[] =
 "Kodak P850",
 "Kodak P880",
 "Kodak Z980",
+"Kodak Z981",
 "Kodak Z1015",
 "Kodak KAI-0340",
 "Konica KD-400Z",
@@ -1756,6 +2065,7 @@ static const char  *static_camera_list[] =
 "Nikon D2X",
 "Nikon D2Xs",
 "Nikon D3",
+"Nikon D3s",
 "Nikon D3X",
 "Nikon D40",
 "Nikon D40X",
@@ -1771,7 +2081,9 @@ static const char  *static_camera_list[] =
 "Nikon D300s",
 "Nikon D700",
 "Nikon D3000",
+"Nikon D3100",
 "Nikon D5000",
+"Nikon D7000",
 "Nikon E700 (\"DIAG RAW\" hack)",
 "Nikon E800 (\"DIAG RAW\" hack)",
 "Nikon E880 (\"DIAG RAW\" hack)",
@@ -1792,8 +2104,10 @@ static const char  *static_camera_list[] =
 "Nikon E8700",
 "Nikon E8800",
 "Nikon Coolpix P6000",
+"Nikon Coolpix P7000",
 "Nikon Coolpix S6 (\"DIAG RAW\" hack)",
 "Nokia N95",
+"Nokia X2",
 "Olympus C3030Z",
 "Olympus C5050Z",
 "Olympus C5060WZ",
@@ -1805,6 +2119,7 @@ static const char  *static_camera_list[] =
 "Olympus X200,D560Z,C350Z",
 "Olympus E-1",
 "Olympus E-3",
+"Olympus E-5",
 "Olympus E-10",
 "Olympus E-20",
 "Olympus E-30",
@@ -1818,6 +2133,8 @@ static const char  *static_camera_list[] =
 "Olympus E-520",
 "Olympus E-620",
 "Olympus E-P1",
+"Olympus E-P2",
+"Olympus E-PL1",
 "Olympus SP310",
 "Olympus SP320",
 "Olympus SP350",
@@ -1831,16 +2148,24 @@ static const char  *static_camera_list[] =
 "Panasonic DMC-FZ28",
 "Panasonic DMC-FZ30",
 "Panasonic DMC-FZ35/FZ38",
+"Panasonic DMC-FZ40",
 "Panasonic DMC-FZ50",
+"Panasonic DMC-FZ100",
 "Panasonic DMC-FX150",
 "Panasonic DMC-G1",
+"Panasonic DMC-G10",
+"Panasonic DMC-G2",
+"Panasonic DMC-GF1",
+"Panasonic DMC-GF2",
 "Panasonic DMC-GH1",
+"Panasonic DMC-GH2",
 "Panasonic DMC-L1",
 "Panasonic DMC-L10",
 "Panasonic DMC-LC1",
 "Panasonic DMC-LX1",
 "Panasonic DMC-LX2",
 "Panasonic DMC-LX3",
+"Panasonic DMC-LX5",
 "Pentax *ist D",
 "Pentax *ist DL",
 "Pentax *ist DL2",
@@ -1853,11 +2178,14 @@ static const char  *static_camera_list[] =
 "Pentax K200D",
 "Pentax K2000/K-m",
 "Pentax K-x",
+"Pentax K-r",
+"Pentax K-5",
 "Pentax K-7",
 "Pentax Optio S",
 "Pentax Optio S4",
 "Pentax Optio 33WR",
 "Pentax Optio 750Z",
+"Pentax 645D",
 "Phase One LightPhase",
 "Phase One H 10",
 "Phase One H 20",
@@ -1867,14 +2195,29 @@ static const char  *static_camera_list[] =
 "Phase One P 30",
 "Phase One P 45",
 "Phase One P 45+",
+"Phase One P 65",
 "Pixelink A782",
+#ifdef LIBRAW_DEMOSAIC_PACK_GPL2
+"Polaroid x530",
+#endif
 "Rollei d530flex",
 "RoverShot 3320af",
+"Samsung EX1",
 "Samsung GX-1S",
-"Samsung GX-10",
+"Samsung GX10",
+"Samsung GX20",
+"Samsung NX10",
+"Samsung NX100",
+"Samsung WB550",
+"Samsung WB2000",
 "Samsung S85 (hacked)",
 "Samsung S850 (hacked)",
 "Sarnoff 4096x5440",
+#ifdef LIBRAW_DEMOSAIC_PACK_GPL2
+"Sigma SD9",
+"Sigma SD10",
+"Sigma SD14",
+#endif
 "Sinar 3072x2048",
 "Sinar 4080x4080",
 "Sinar 4080x5440",
@@ -1891,14 +2234,19 @@ static const char  *static_camera_list[] =
 "Sony DSLR-A330",
 "Sony DSLR-A350",
 "Sony DSLR-A380",
+"Sony DSLR-A450",
 "Sony DSLR-A500",
 "Sony DSLR-A550",
+"Sony DSLR-A580",
 "Sony DSLR-A700",
 "Sony DSLR-A850",
 "Sony DSLR-A900",
+"Sony NEX-3",
+"Sony NEX-5",
+"Sony SLT-A33",
+"Sony SLT-A55V",
 "Sony XCD-SX910CR",
 "STV680 VGA",
-
    NULL
 };
 
@@ -1926,6 +2274,8 @@ const char * LibRaw::strprogress(enum LibRaw_progress p)
             return "Removing dead pixels";
         case LIBRAW_PROGRESS_DARK_FRAME:
             return "Subtracting dark frame data";
+        case LIBRAW_PROGRESS_FOVEON_INTERPOLATE:
+            return "Interpolating Foveon sensor data";
         case LIBRAW_PROGRESS_SCALE_COLORS:
             return "Scaling colors";
         case LIBRAW_PROGRESS_PRE_INTERPOLATE:
@@ -1954,3 +2304,8 @@ const char * LibRaw::strprogress(enum LibRaw_progress p)
             return "Some strange things";
         }
 }
+
+
+
+
+
